;;; This source file is part of the Meta-CVS program, 
;;; which is distributed under the GNU license.
;;; Copyright 2002 Kaz Kylheku

(require "system")
(require "dirwalk")
(require "chatter")
(require "split")
(require "mapping")
(require "rcs-utils")
(provide "convert")

(defun remove-attic-component (path)
  (let* ((split-path (split-fields path "/"))
         (len (length split-path)))
    (when (and (>= len 2) (string= (nth (- len 2) split-path) "Attic"))
      (setf split-path (append (butlast (butlast split-path)) 
			       (last split-path))))
    (reduce #'(lambda (x y) (format nil "~a/~a" x y)) split-path)))

(defun classify-tags (tags)
  (let (version-tags branch-tags)
    (dolist (tag tags (values version-tags branch-tags))
      (destructuring-bind (tag-name tag-value) tag
	(if (search ".0." tag-value)
	  (push tag-name branch-tags)
	  (push tag-name version-tags))))))
	
(defun mcvs-convert (source-dir target-dir)
  (when (ignore-errors (stat target-dir))
    (error "a directory or file called ~a exists here already."
           target-dir))

  (multiple-value-bind (path created) 
                       (ensure-directories-exist (path-cat target-dir 
							   *mcvs-map-name*))
    (declare (ignore path))
    (if (not created) 
      (error "unable to create directory ~a." target-dir)))

  (let (filemap all-version-tags all-branch-tags)
    (current-dir-restore
      (chdir source-dir)
      (for-each-file-info (fi ".")
	(when (regular-p fi)
	  (let ((canon-name (canonicalize-path (file-name fi))))
	    (multiple-value-bind (suffix v-suffix-free-name) 
				 (suffix canon-name #\,)
	      (when (and suffix (string= suffix "v"))
		(let ((attic-free-name 
			(remove-attic-component v-suffix-free-name))
		      (f-name (mapping-generate-id :suffix (suffix v-suffix-free-name)
						   :no-dir t)))
		  (link canon-name (path-cat *up-dir* 
					     target-dir 
					     (format nil "~A,v" f-name)))
		  (push (make-mapping-entry :kind :file
					    :id (path-cat *mcvs-dir* f-name)
					    :path attic-free-name
					    :executable (executable-p fi))
			filemap)
		  (with-open-file (f (parse-posix-namestring canon-name) 
				     :direction :input)
		    (chatter-info "scanning ~a.~%" canon-name)
		    (let ((rcs-file (rcs-parse f)))
		      (multiple-value-bind (version-tags branch-tags)
					   (classify-tags (rcs-admin-symbols (rcs-file-admin rcs-file)))
			(setf all-version-tags (nunion all-version-tags 
						       version-tags 
						       :test #'string=))
			(setf all-branch-tags (nunion all-branch-tags branch-tags 
						      :test #'string=))))))))))))

    (current-dir-restore
      (chdir target-dir)
      (chatter-info "Writing map.")
      (mapping-write filemap *mcvs-map-name*)

      (execute-program `("ci" "-mMeta-CVS MAP file created by mcvs convert."
			 "-t/dev/null" ,*mcvs-map-name*))
      (execute-program `("chmod" "ug+rw" ,(format nil "~A,v" *mcvs-map-name*)))

      (execute-program-xargs '("rcs") 
			     (mapcar #'(lambda (tag)
					 (format nil "-n~A:1.1" tag))
				     all-version-tags)
			     (list *mcvs-map-name*))

      (let ((branch-counter 0))
	(execute-program-xargs '("rcs")
			       (mapcar #'(lambda (tag)
					   (format nil 
						   "-n~A:1.1.0.~A" 
						   tag (+ 2 branch-counter)))
				       all-branch-tags)
			       (list *mcvs-map-name*))))))

(defun mcvs-convert-wrapper (cvs-options cvs-command-options mcvs-args)
  (declare (ignore cvs-options cvs-command-options))
  (if (/= (length mcvs-args) 2)
    (error "specify cvs source dir and new target dir."))
  (mcvs-convert (first mcvs-args) (second mcvs-args)))
